using System;
using OpenTK.Graphics.OpenGL;

namespace Game
{
    /// <summary>
    /// A simple GLSL shader class.
    /// </summary>
    class GLSLShader
    {
        #region -- Fields --
        /// <summary>
        /// The GLint of the shader program after compile/link.
        /// </summary>
        private int programID;
        #endregion

         
        #region -- Properties --
        /// <summary>
        /// Gets the program ID.
        /// </summary>
        public int ProgramID
        {
            get
            {
                return programID;
            }
        } 
        #endregion


        #region -- Constructor -- 
        /// <summary>
        /// Initializes a new instance of the <see cref="GLSLShader"/> class.
        /// </summary>
        /// <param name="name">The filename of the shader to parse and compile.</param>
        public GLSLShader(string name)
        {
            BuildShader(name);
        } 
        #endregion


        #region BuildShader
        /// <summary>
        /// Parses and builds the shader.
        /// </summary>
        /// <param name="name">The filename of the shader to parse and compile.</param>
        /// <returns></returns>
        private int BuildShader(string name)
        {
            // Send the shader to the preprocessor to be chopped
            // and have its includes processed.
            var shaderSource = new ShaderPreprocessor(name);

            // declare locals to hold glStates
            int vertexShader, fragmentShader, statusCode;
            string info;

            vertexShader = GL.CreateShader(ShaderType.VertexShader);
            fragmentShader = GL.CreateShader(ShaderType.FragmentShader);

            #region Compile Vertex
            // Compile the vertex shader or report error on failure.
            GL.ShaderSource(vertexShader, shaderSource.VertexShaderSource);
            GL.CompileShader(vertexShader);
            GL.GetShaderInfoLog(vertexShader, out info);
            GL.GetShader(vertexShader, ShaderParameter.CompileStatus, out statusCode);

            if (statusCode != 1)
            {
                Console.WriteLine(shaderSource.VertexShaderSource);
                throw new ApplicationException(info);
            }
            else Console.WriteLine("Shader::Vertex shader {0} Compiled properly.", name); 
            #endregion

            #region Compile Fragment
            // Compile the fragment shader or report error on failure.
            GL.ShaderSource(fragmentShader, shaderSource.FragmentShaderSource);
            GL.CompileShader(fragmentShader);
            GL.GetShaderInfoLog(fragmentShader, out info);
            GL.GetShader(fragmentShader, ShaderParameter.CompileStatus, out statusCode);

            if (statusCode != 1)
                 throw new ApplicationException(info);
            else Console.WriteLine("Shader::Fragment shader {0} Compiled properly.", name); 
            #endregion

            #region Link
            // Link the program or report error on failure.
            programID = GL.CreateProgram();
            GL.AttachShader(programID, fragmentShader);
            GL.AttachShader(programID, vertexShader);
            GL.LinkProgram(programID);

            GL.GetProgramInfoLog(programID, out info);
            GL.GetProgram(programID, ProgramParameter.LinkStatus, out statusCode);
            if (statusCode != 1)
                throw new ApplicationException(info);
            else Console.WriteLine("Shader::{0} linked properly.", name); 
            #endregion
            return programID;
        } // BuildShader()
        #endregion


        #region Bind
        /// <summary>
        /// Binds this shader to be ther active one.
        /// </summary>
        public void Bind()
        {
            GL.UseProgram(programID);
        }
        #endregion


        #region UnBind
        /// <summary>
        /// Bind the default shader.
        /// </summary>
        public void UnBind()
        {
            GL.UseProgram(0);
        }
        #endregion
    }
}
