
//#define SHADER_DEBUG

using System;
using System.Collections.Generic;
using System.IO;

namespace Game
{
    /// <summary>
    /// Open and process a GLSL shader. This preprocessor allows the tags:
    ///  * #pragma vertex_shader - The start of the vertex shader.
    ///  * #pragma fragment_shader - The start of the fragment shader.
    ///  * #pragma include [filename] - A file to include in place.
    /// </summary>
    public class ShaderPreprocessor
    {
        #region -- Fields -- 
        string vertexShader;
        string fragmentShader;
        Logging log = Logging.Instance;
        #endregion
        
        
        #region -- Constructor --
        /// <summary>
        /// Initializes a new instance of the <see cref="ShaderPreprocessor"/> class.
        /// </summary>
        /// <param name="filename">The filename of the shader to split and parse.</param>
        public ShaderPreprocessor (string filename)
        {
            LoadShaderFile(filename);
            #if SHADER_DEBUG
                log.Info( string.Format("ShaderPreprocessor:: {0}", filename) );
                log.Info(vertexShader);
                log.Info("=====================================================");
                log.Info(fragmentShader);
            #endif

        }
        #endregion
        
        
        #region Properties
        /// <summary>
        /// Gets the vertex shader source as a string suitable for compilation.
        /// </summary>
        public string VertexShaderSource
        {
            get
            {
                return vertexShader;
            }
        }


        /// <summary>
        /// Gets the fragment shader source as a string suitable for compilation.
        /// </summary>
        public string FragmentShaderSource
        {
            get
            {
                return fragmentShader;
            }
        }
        #endregion
        
        
        #region SplitShaders
        /// <summary>
        /// Splits the shaders.
        /// </summary>
        /// <param name="shaderBuffer">The shader buffer.</param>
        private void SplitShaders(List<string> shaderBuffer){
            int vertexShaderOffset=0, fragmentShaderOffset=0, lineCount=0;
            foreach (string line in shaderBuffer)
            {
                if (line == "#pragma vertex_shader"){
                    vertexShaderOffset = lineCount+1;
                }
                else if (line == "#pragma fragment_shader"){
                    fragmentShaderOffset = lineCount+1;
                }
                lineCount++;
            }
            List<string> vertexShaderBuffer = 
                shaderBuffer.GetRange(vertexShaderOffset, fragmentShaderOffset-vertexShaderOffset-2);
            List<string> fragmentShaderBuffer = 
                shaderBuffer.GetRange(fragmentShaderOffset, lineCount-fragmentShaderOffset);
            ProcessIncludes(vertexShaderBuffer, out vertexShader);
            ProcessIncludes(fragmentShaderBuffer, out fragmentShader);
        }
        #endregion
        
        
        #region LoadShaderFile
        /// <summary>
        /// Loads the shader file from the disk.
        /// </summary>
        /// <param name="filename">The filename.</param>
        private void LoadShaderFile(string filename)
        {
            List<string> shaderStringBuffer = new List<string>();
            using (StreamReader shaderFile = new StreamReader(string.Format("./{0}", filename)))
            while(shaderFile.Peek() >= 0)
            {
                shaderStringBuffer.Add(shaderFile.ReadLine());
            }
            SplitShaders(shaderStringBuffer);
        }
        #endregion
            
        
        #region ProcessIncludes
        /// <summary>
        /// Processes the includes.
        /// </summary>
        /// <param name="shaderBuffer">The shader buffer.</param>
        /// <param name="shaderOut">The shader out.</param>
        private void ProcessIncludes(List<string> shaderBuffer, out string shaderOut)
        {
            string finalShader = string.Empty;
            foreach (string line in shaderBuffer)
            {
                if (line.Contains("#pragma include")){
                    var splitLine = line.Split(' ');
                    string includeFilename = string.Format("{0}",splitLine[2]);
                    finalShader += string.Format("// included from source file {0}\n\r", includeFilename);
                    using (StreamReader includeFileReader = new StreamReader(includeFilename)){
                        finalShader += includeFileReader.ReadToEnd();    
                    }
                }
                else
                {
                    finalShader += line + "\n";
                }
            }
            shaderOut = finalShader;
        }
        #endregion
        

    }
}

